package org.pelds;

import java.io.File;
import java.util.Properties;
import java.util.logging.Logger;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.util.FileManager;
import com.hp.hpl.jena.util.FileUtils;

/**
 * Interface to the PeLDS configuration variables
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class Config {

	// configuration settings
	private String dbUrl;
	private String dbUser;
	private String dbPass;
	private String dbDriver;

	private String keystoreLocation;
	private String keystorePass;

	private String serverCertName;
	private String serverCertPass;

	private String dereferencingPrefix;
	private String dereferencingEndpoint;

	private String prefixMapFile;

	// NOTE: if you add a variable above, also add a suitable loading call
	// within init()

	private String configDir;
	private PrefixMapping prefixMapping = null;
	// configuration file related
	private static String KEY_PREFIX = "pelds.";
	// initialization
	static private Logger log = Logger.getLogger("Config");

	/**
	 * Constructor for default values
	 * 
	 */
	public Config() {
		this.configDir = ".";
		init(new Properties());
	}

	/**
	 * Initialize configuration variables from Properties or by default value
	 * 
	 * @param p
	 *            Properties object loaded from the configuration file by
	 *            servlets
	 * @see org.pelds.servlet.Servlet
	 */
	public void init(Properties p) {
		this.dbUrl = p.getProperty(KEY_PREFIX + "db.url",
				"jdbc:mysql://localhost/pelds");
		this.dbUser = p.getProperty(KEY_PREFIX + "db.user", "pelds");
		this.dbPass = p.getProperty(KEY_PREFIX + "db.pass", "pelds");
		this.dbDriver = p.getProperty(KEY_PREFIX + "db.driver",
				"com.mysql.jdbc.Driver");

		this.keystoreLocation = addAbsolutePath(p.getProperty(KEY_PREFIX
				+ "cert.keystore.location", "peldskeystore.jks"));
		this.keystorePass = p.getProperty(KEY_PREFIX + "cert.keystore.pass",
				"pelds");

		this.serverCertName = p.getProperty(KEY_PREFIX + "cert.server.name",
				"pelds");
		this.serverCertPass = p.getProperty(KEY_PREFIX + "cert.server.pass",
				"pelds");

		this.dereferencingPrefix = p.getProperty(KEY_PREFIX
				+ "dereferencing.prefix", "");
		this.dereferencingEndpoint = p.getProperty(KEY_PREFIX
				+ "dereferencing.endpoint",
				"https://localhost:8443/pelds/query");

		this.prefixMapFile = addAbsolutePath(p.getProperty(KEY_PREFIX
				+ "prefixes.map", "prefixes.n3"));
		// NOTE: add load call and default value for additional configuration
		// parameters above
	}

	/**
	 * Add a predefined path to file configuration values in order to enable
	 * paths relative to the configuration file
	 * 
	 * @param fileName
	 *            a file name
	 * @return absolute path if found
	 */
	private String addAbsolutePath(String fileName) {
		File f = new File(fileName);
		if (!f.exists())
			f = new File(getConfigDir() + File.separator + fileName);
		return f.getAbsolutePath();
	}

	/**
	 * Load prefix map from RDF serialization
	 * 
	 * @return PrefixMapping from file
	 */
	public PrefixMapping getPrefixMap() {
		if (this.prefixMapping != null)
			return this.prefixMapping;
		String mapFile = getPrefixMapFile();
		log.info("Using prefix map " + mapFile);
		Model m = ModelFactory.createDefaultModel();
		String lang = FileUtils.guessLang(mapFile, "N3");
		try {
			FileManager.get().readModel(m, mapFile, lang);
		} catch (Exception e) {
			log.warning("Failed to use prefix map " + mapFile);
		}
		PrefixMapping p = PrefixMapping.Factory.create();
		p.setNsPrefixes(m.getNsPrefixMap());
		this.prefixMapping = p;
		return p;
	}

	/**
	 * @param configDir
	 */
	public void setConfigDir(String configDir) {
		this.configDir = configDir;
	}

	public String getDbUrl() {
		return dbUrl;
	}

	public String getDbUser() {
		return dbUser;
	}

	public String getDbPass() {
		return dbPass;
	}

	public String getDbDriver() {
		return dbDriver;
	}

	public String getServerCertName() {
		return serverCertName;
	}

	public String getServerCertPass() {
		return serverCertPass;
	}

	public String getKeystoreLocation() {
		return keystoreLocation;
	}

	public String getKeystorePass() {
		return keystorePass;
	}

	public String getDereferencingPrefix() {
		return dereferencingPrefix;
	}

	public String getDereferencingEndpoint() {
		return dereferencingEndpoint;
	}

	public String getConfigDir() {
		return configDir;
	}

	public String getPrefixMapFile() {
		return prefixMapFile;
	}

}