package org.pelds.auth;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.net.URL;
import java.security.Key;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.logging.Logger;

import org.bouncycastle.asn1.misc.MiscObjectIdentifiers;
import org.bouncycastle.asn1.misc.NetscapeCertType;
import org.bouncycastle.asn1.x509.AuthorityKeyIdentifier;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.x509.KeyUsage;
import org.bouncycastle.asn1.x509.SubjectKeyIdentifier;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.jce.netscape.NetscapeCertRequest;
import org.bouncycastle.util.encoders.Base64;
import org.bouncycastle.x509.X509V3CertificateGenerator;
import org.bouncycastle.x509.extension.AuthorityKeyIdentifierStructure;
import org.bouncycastle.x509.extension.SubjectKeyIdentifierStructure;
import org.pelds.Config;

import com.hp.hpl.jena.sparql.util.Utils;

/**
 * X509 certificate signing with custom Subject Alt Name values
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class CreateCert {
	static private Logger log = Logger.getLogger("CreateCert");

	private static KeyStore keystore = null;

	private String serverCertName;
	private String serverCertPw;
	private String keystorePw;
	private String keystoreLocation;

	private X509Certificate cert = null;

	static {
		Security
				.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
	}

	/**
	 * Constructor
	 * 
	 * @param c
	 *            a PeLDS configuration object
	 * @see org.pelds.Config
	 */
	public CreateCert(Config c) {
		keystoreLocation = c.getKeystoreLocation();
		keystorePw = c.getKeystorePass();
		serverCertName = c.getServerCertName();
		serverCertPw = c.getServerCertPass();
	}

	/**
	 * Create the certificate from a "Netscape signed public key and challenge"
	 * (SPKAC)
	 * 
	 * @param lifetimeDays
	 *            Certificate lifetime in days
	 * @param cn
	 *            Common Name (CN) for certificate
	 * @param certURL
	 *            Subject Alternative Name as Web ID
	 * @param spkac
	 *            the encoded SPKAC
	 * @return a new X509 Certificate
	 * @throws Exception
	 * @see http://openssl.org/docs/apps/spkac.html
	 */
	public X509Certificate createCert(int lifetimeDays, String cn, URL certURL,
			String spkac) throws Exception {

		if (cert != null)
			return cert;
		KeyStore ks = getKeyStore();
		if (ks == null) {
			throw new Exception();
		}
		X509Certificate serverCert = (X509Certificate) ks
				.getCertificate(serverCertName);
		log.info("cert=" + serverCert);

		Key serverKey = ks.getKey(serverCertName, serverCertPw.toCharArray());
		NetscapeCertRequest netsCert = getCertReq(spkac);
		log.info("cert req challenge=" + netsCert.getChallenge());
		log.info("cert req public key=" + netsCert.getPublicKey());
		X509V3CertificateGenerator certGenerator = new X509V3CertificateGenerator();
		certGenerator.reset();
		/*
		 * The serial-number of this certificate
		 */
		// todo: put a counter here and add some randomness sha1(secret + time +
		// something)
		// in order to bypass the problem of
		// http://www.win.tue.nl/hashclash/rogue-ca/
		BigInteger num = BigInteger.valueOf(new Date().getTime());
		certGenerator.setSerialNumber(num);
		certGenerator.setSubjectDN(new X509Name(
				"C=DE, O=PeLDS Certificates, OU=" + num + ",CN=" + cn));
		certGenerator.setIssuerDN(serverCert.getIssuerX500Principal());
		certGenerator.setPublicKey(netsCert.getPublicKey());
		/*
		 * Sets the signature algorithm.
		 */
		String pubKeyAlgorithm = serverCert.getPublicKey().getAlgorithm();
		if (pubKeyAlgorithm.equals("DSA")) {
			certGenerator.setSignatureAlgorithm("SHA1WithDSA");
		} else if (pubKeyAlgorithm.equals("RSA")) {
			certGenerator.setSignatureAlgorithm("SHA1WithRSAEncryption");
		} else {
			RuntimeException re = new RuntimeException(
					"Algorithm not recognised: " + pubKeyAlgorithm);
			// LOGGER.error(re.getMessage(), re);
			throw re;
		}
		Date now = new Date();
		certGenerator.setNotBefore(now);
		certGenerator.setNotAfter(new Date(now.getTime()
				+ ((long) lifetimeDays * 24L * 60L * 60L * 1000L))); // 1 year
		/*
		 * Adds the Basic Constraint (CA: true) extension.
		 */
		certGenerator.addExtension(X509Extensions.BasicConstraints, true,
				new BasicConstraints(false));
		/*
		 * Adds the Key Usage extension.
		 */
		certGenerator.addExtension(X509Extensions.KeyUsage, true, new KeyUsage(
				KeyUsage.digitalSignature | KeyUsage.nonRepudiation
						| KeyUsage.keyEncipherment | KeyUsage.keyAgreement
						| KeyUsage.keyCertSign));
		/*
		 * Adds the Netscape certificate type extension.
		 */
		certGenerator.addExtension(MiscObjectIdentifiers.netscapeCertType,
				false, new NetscapeCertType(NetscapeCertType.sslClient
						| NetscapeCertType.smime));
		/*
		 * Adds the subject key identifier extension.
		 */
		SubjectKeyIdentifier subjectKeyIdentifier = new SubjectKeyIdentifierStructure(
				netsCert.getPublicKey());
		certGenerator.addExtension(X509Extensions.SubjectKeyIdentifier, false,
				subjectKeyIdentifier);
		/*
		 * Adds the authority key identifier extension.
		 */
		AuthorityKeyIdentifier authorityKeyIdentifier = new AuthorityKeyIdentifierStructure(
				serverCert.getPublicKey());
		certGenerator.addExtension(X509Extensions.AuthorityKeyIdentifier,
				false, authorityKeyIdentifier);
		/*
		 * Adds the subject alternative-name extension.
		 */
		// if (subjAltNameURI != null) {
		GeneralNames subjectAltNames = new GeneralNames(new GeneralName(
				GeneralName.uniformResourceIdentifier, certURL.toString()));
		certGenerator.addExtension(X509Extensions.SubjectAlternativeName,
				false, subjectAltNames);
		// }

		/*
		 * Creates and sign this certificate with the private key corresponding
		 * to the public key of the certificate (hence the name
		 * "self-signed certificate").
		 */
		X509Certificate returncert = certGenerator
				.generate((PrivateKey) serverKey);
		/*
		 * Checks that this certificate has indeed been correctly signed.
		 */
		returncert.verify(serverCert.getPublicKey());

		log.info("return cert=" + returncert);
		cert = returncert;
		return returncert;
	}

	/**
	 * Open a Java Key Store containing the server certificate
	 * 
	 * @return Java Key Store instance
	 */
	private KeyStore getKeyStore() {
		if (keystore != null)
			return keystore;

		try {
			log.info("Opening the keystore in " + keystoreLocation);

			keystore = KeyStore.getInstance("JKS");
			InputStream ksInputStream = new FileInputStream(keystoreLocation);
			keystore.load(ksInputStream, keystorePw.toCharArray());
			ksInputStream.close();

		} catch (Exception e) {
			log
					.warning("Opening the keystore in " + keystoreLocation
							+ " caused a " + Utils.className(e) + ": "
							+ e.getMessage());
		}
		return keystore;
	}

	/**
	 * Decode the SPKAC
	 * 
	 * @param spkac
	 * @return NetscapeCertRequest instance
	 * @throws IOException
	 */
	NetscapeCertRequest getCertReq(String spkac) throws IOException {
		byte[] bts = Base64.decode(spkac);
		return new NetscapeCertRequest(bts);
	}

}
