package org.pelds.auth;

import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;

import net.java.dev.sommer.foafssl.principals.FoafSslPrincipal;
import net.java.dev.sommer.foafssl.verifier.FoafSslVerifier;

import com.hp.hpl.jena.sparql.util.Utils;

/**
 * Interface to FOAF+SSL authentication
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class FoafSslInterface {
	static private Logger log = Logger.getLogger("FoafSslInterface");

	private static HashMap<String, Collection<String>> certificateCache = new HashMap<String, Collection<String>>();

	/**
	 * Do FOAF+SSL authentication from SSL client certificates
	 * 
	 * @param req
	 *            HTTP request containing client certificates
	 * @return List of authenticated URLs
	 */
	public static Collection<String> getVerifiedWebIds(HttpServletRequest req) {
		Collection<String> verifiedWebIDs = new ArrayList<String>();
		X509Certificate[] certificates = (X509Certificate[]) req
				.getAttribute("javax.servlet.request.X509Certificate");
		if (certificates == null)
			return verifiedWebIDs;

		FoafSslVerifier FOAF_SSL_VERIFIER = new TolerantFoafSslVerifier();

		for (int i = 0; i < certificates.length; i++) {
			X509Certificate foafSslCertificate = certificates[i];
			try {
				Collection<String> curWebIds = new ArrayList<String>();

				String certFingerprint = foafSslCertificate.toString();

				// load uri list from cache if possible
				if (certificateCache.containsKey(certFingerprint)) {
					verifiedWebIDs
							.addAll(certificateCache.get(certFingerprint));
					continue;
				}

				Iterator<? extends FoafSslPrincipal> verifiedWebIDsForCert = FOAF_SSL_VERIFIER
						.verifyFoafSslCertificate(foafSslCertificate)
						.iterator();

				while (verifiedWebIDsForCert.hasNext()) {
					FoafSslPrincipal fsp = verifiedWebIDsForCert.next();
					curWebIds.add(fsp.getUri().toString());
				}
				verifiedWebIDs.addAll(curWebIds);

				// add verified web ids to runtime cache, make copy of uri list
				Collection<String> cacheIdList = new ArrayList<String>();
				cacheIdList.addAll(curWebIds);
				if (curWebIds.size() > 0)
					certificateCache.put(certFingerprint, cacheIdList);

			} catch (Exception e) {
				log.warning("Authentication caused a " + Utils.className(e)
						+ ": " + e.getMessage());
				e.printStackTrace();
			}
		}
		return verifiedWebIDs;
	}
}
