package org.pelds.auth;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.security.PublicKey;
import java.util.logging.Logger;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import net.java.dev.sommer.foafssl.principals.DereferencedFoafSslPrincipal;
import net.java.dev.sommer.foafssl.verifier.DereferencingFoafSslVerifier;

import org.openrdf.OpenRDFException;

/**
 * This small derived class overwrites some ssl functions to ignore crappy
 * certificates and does proper redirecting, even from HTTP to HTTPS URLs. DOH!
 * Certificate errors can be ignored as only the encryption is wanted
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class TolerantFoafSslVerifier extends DereferencingFoafSslVerifier {

	static transient Logger log = Logger
			.getLogger(TolerantFoafSslVerifier.class.getName());

	/**
	 * How many redirects should this class follow?
	 */
	private static int REDIRECT_LIMIT = 4;

	// set up a dummy "TrustManager" to ignore certificate chain errors of any
	// kind
	private static TrustManager[] trustAllCerts = new TrustManager[] { new X509TrustManager() {
		public java.security.cert.X509Certificate[] getAcceptedIssuers() {
			return null;
		}

		public void checkClientTrusted(
				java.security.cert.X509Certificate[] certs, String authType) {
		}

		public void checkServerTrusted(
				java.security.cert.X509Certificate[] certs, String authType) {
		}
	} };
	
	private static HostnameVerifier hv = new HostnameVerifier() {
		public boolean verify(String arg0, SSLSession arg1) {
			return true;
		}
	};

	/**
	 * Constructor, disables - Host name - Certificate common name matching -
	 * SSL certificate trust checking
	 */
	public TolerantFoafSslVerifier() {
		// set up a dummy "HostnameVerifier" to ignore host name / certificate
		// common
		// name mismatches
		HttpsURLConnection.setDefaultHostnameVerifier(hv);

		// ignore certificate chain errors
		SSLSocketFactory factory = (SSLSocketFactory) SSLSocketFactory
				.getDefault();
		try {
			SSLContext sc = SSLContext.getInstance("SSL");
			sc.init(null, trustAllCerts, new java.security.SecureRandom());
			factory = sc.getSocketFactory();
		} catch (Exception e) {
			// silently fail
		}
		HttpsURLConnection.setDefaultSSLSocketFactory(factory);
	}

	/**
	 * Overwrite the central function to do those redirection between protocols
	 * URLConnection ignores.
	 * 
	 * @see net.java.dev.sommer.foafssl.verifier.DereferencingFoafSslVerifier
	 */
	public DereferencedFoafSslPrincipal verifyByDereferencing(URI claimedIdUri,
			PublicKey certPublicKey) throws OpenRDFException, IOException {

		URL foafname = claimedIdUri.toURL();
		URL actualURL;
		URLConnection conn;

		String redirect = foafname.toString();
		int redirectCount = 0;

		do {
			actualURL = new URL(redirect);
			conn = actualURL.openConnection();

			conn.addRequestProperty("Accept:", "application/rdf+xml");
			conn.connect();
			redirect = conn.getHeaderField("Location");
			redirectCount++;

		} while (redirect != null && redirectCount < REDIRECT_LIMIT);

		if (redirectCount > 1)
			log.info("Manually resolved <" + foafname.toString() + "> to <"
					+ actualURL.toString() + ">");
		InputStream is = conn.getInputStream();
		try {
			String mimeType = mimeType(conn.getContentType());
			// dereferencedSecurely cannot be true anymore, i guess
			return verifyByDereferencing(claimedIdUri, certPublicKey, conn
					.getURL(), is, mimeType, false, null);
		} finally {
			is.close();
		}
	}

	/**
	 * Retrieve mime type
	 * 
	 * @param contentType Content type from HTTP header
	 * @return a mime type
	 */
	private String mimeType(String contentType) {
		int i = contentType.indexOf(';');
		if (i > 0) {
			contentType = contentType.substring(0, i);
		}
		return contentType.trim();
	}
}