package org.pelds.rules;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.mindswap.pellet.jena.vocabulary.SWRL;
import org.mindswap.pellet.utils.Namespaces;

import com.hp.hpl.jena.datatypes.xsd.XSDDatatype;
import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;
import com.sun.org.apache.xml.internal.utils.URI;

/**
 * Transforms rules from text representation into the internal object structure
 * and into RDF/SWRL
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class RuleParser {
	private Logger log = Logger.getLogger("RuleParser");
	private ArrayList<Rule> ruleList = new ArrayList<Rule>();
	private PrefixMapping prefixMapping;
	private OntModel model;
	private SWRLFactory swrl;
	private Model ruleModel;

	/**
	 * Constructor, takes rules in their text representation
	 * 
	 * @param rules
	 *            rules in text representation
	 * @param verbose
	 *            debug flag
	 */
	public RuleParser(String rules, boolean verbose) {
		if (rules == null)
			return;
		setVerbose(verbose);

		Iterator<Map.Entry<String, String>> rawIt = findRules(
				removeComments(rules)).entrySet().iterator();
		while (rawIt.hasNext()) {
			Rule r = parseRule(rawIt.next());
			if (r != null)
				ruleList.add(r);
		}
	}

	/**
	 * Finds rules in a String by using a regex
	 * 
	 * @param rules
	 *            String containing rules
	 * @return Map from rule name to rule content
	 */
	private Map<String, String> findRules(String rules) {
		HashMap<String, String> map = new HashMap<String, String>();
		Matcher m = Pattern.compile("(\\w+):([^;]+);", Pattern.DOTALL).matcher(
				rules);

		while (m.find()) {
			String name = m.group(1).trim();
			String body = m.group(2).trim();
			map.put(name, body);
		}
		return map;
	}

	/**
	 * Finds atoms from a rule content in text form
	 * 
	 * @param rawRule
	 *            rule content
	 * @return Rule instance
	 * @see Rule
	 */
	private Rule parseRule(Map.Entry<String, String> rawRule) {
		String ruleName = rawRule.getKey();
		log.info("Parsing rule " + ruleName);
		Matcher m = Pattern.compile("(.+)?=>(.+)", Pattern.DOTALL).matcher(
				rawRule.getValue());
		if (!m.matches()) {
			log.info("Failed to parse rule " + ruleName);
			return null;
		}
		String antecedent = m.group(1);
		if (antecedent == null) antecedent = "";

		antecedent = antecedent.trim();
		String consequent = m.group(2).trim();

		Rule r = new Rule();
		r.setAntAtoms(findAtoms(antecedent));
		r.setConAtoms(findAtoms(consequent));
		r.setName(ruleName);
		
		log.info("Result: " + r.toString());
		return r;
	}

	/**
	 * Parses atoms from text form
	 * 
	 * @param atomListStr
	 *            several atoms in text form
	 * @return List of RuleAtom instances
	 * @see RuleAtom
	 */
	private ArrayList<RuleAtom> findAtoms(String atomListStr) {
		Matcher m = Pattern.compile("([\\w/:@#-_]+)\\(([^)]*)\\)",
				Pattern.DOTALL).matcher(atomListStr);
		ArrayList<RuleAtom> atomList = new ArrayList<RuleAtom>();
		while (m.find()) {
			String name = m.group(1).trim();
			String params = m.group(2).trim();
			RuleAtom r = new RuleAtom();
			r.setName(name);

			StringTokenizer st = new StringTokenizer(params, ",");
			while (st.hasMoreTokens()) {
				String p = st.nextToken().trim();
				if (!p.equals(""))
					r.addArgument(p);
			}
			atomList.add(r);
		}
		return atomList;
	}

	/**
	 * Converts Rules from internal model into RDF/SWRL format
	 * 
	 * @param m
	 *            Model to read property types etc from
	 * @param p
	 *            PrefixMapping containing the valid prefixes for atoms
	 * @param actionUri
	 *            URI of current Action
	 * @return RDF Model containing SWRL rules with triple patterns
	 */
	public Model addRulesToModel(OntModel m, PrefixMapping p, String actionUri) {
		prefixMapping = p;
		model = m;
		ruleModel = ModelFactory.createDefaultModel();
		swrl = new SWRLFactory(ruleModel);

		Iterator<Rule> ruleIt = ruleList.iterator();
		// for each rule
		while (ruleIt.hasNext()) {
			Rule r = ruleIt.next();
			String ruleUri = RuleQuery.createRandomizedUri("rule_"
					+ r.getName());
			Resource ruleRes = ruleModel.createResource(ruleUri);

			// for all variables within the rule
			HashMap<String, Resource> varMap = new HashMap<String, Resource>();
			Iterator<String> varIt = r.getUsedVariables().iterator();
			while (varIt.hasNext()) {
				String var = varIt.next();
				Resource varRes = swrl
						.var(RuleQuery.createRandomizedUri("variable_"
								+ var.replace("?", "")));
				varMap.put(var, varRes);
			}
			// add atom resources to model
			ArrayList<Resource> antAtResList = new ArrayList<Resource>();
			Iterator<RuleAtom> antAtIt = r.getAntAtoms().iterator();
			while (antAtIt.hasNext()) {
				antAtResList.add(addRuleAtom(antAtIt.next(), varMap));
			}
			ArrayList<Resource> conAtResList = new ArrayList<Resource>();
			ArrayList<Resource> patternList = new ArrayList<Resource>();

			// TODO: make SWRL variables possible in patterns, may be hard
			Iterator<RuleAtom> conAtIt = r.getConAtoms().iterator();
			while (conAtIt.hasNext()) {
				// remove custom predicates and add patterns
				RuleAtom ra = conAtIt.next();
				String uri = prefixMapping.expandPrefix(ra.getName());

				if (uri.equals(RuleQuery.PELDS_NS + "permit_triple")) {
					// add given triple pattern to rule
					Resource pat = ruleModel.createResource();
					pat.addProperty(RDF.type, model
							.getProperty(RuleQuery.PELDS_NS + "TriplePattern"));
					pat.addProperty(model.getProperty(RuleQuery.PELDS_NS
							+ "tripleSubject"), createParamLiteral(ra
							.getArguments().get(0), varMap));
					pat.addProperty(model.getProperty(RuleQuery.PELDS_NS
							+ "triplePredicate"), createParamLiteral(ra
							.getArguments().get(1), varMap));
					pat.addProperty(model.getProperty(RuleQuery.PELDS_NS
							+ "tripleObject"), createParamLiteral(ra
							.getArguments().get(2), varMap));
					patternList.add(pat);
					continue;
				}

				if (uri.equals(RuleQuery.PELDS_NS + "permit_resource")) {
					// add given resource pattern to rule
					Resource pat = ruleModel.createResource();
					pat.addProperty(RDF.type, model
							.getProperty(RuleQuery.PELDS_NS + "TriplePattern"));
					pat.addProperty(model.getProperty(RuleQuery.PELDS_NS
							+ "tripleSubject"), createParamLiteral(ra
							.getArguments().get(0), varMap));
					pat
							.addProperty(model.getProperty(RuleQuery.PELDS_NS
									+ "triplePredicate"), model
									.createTypedLiteral("*"));
					pat.addProperty(model.getProperty(RuleQuery.PELDS_NS
							+ "tripleObject"), model.createTypedLiteral("*"));
					patternList.add(pat);
					continue;
				}

				if (uri.equals(RuleQuery.PELDS_NS + "permit_instance")) {
					// add given instance pattern to rule
					Resource pat = ruleModel.createResource();
					pat.addProperty(RDF.type, model
							.getProperty(RuleQuery.PELDS_NS + "TriplePattern"));
					pat.addProperty(model.getProperty(RuleQuery.PELDS_NS
							+ "instance"), createParamLiteral(ra.getArguments()
							.get(0), varMap));
					patternList.add(pat);
					continue;
				}
				conAtResList.add(addRuleAtom(ra, varMap));
			}

			// add custom default consequences
			conAtResList.add(swrl.individualPropertyAtom(model
					.getProperty(RuleQuery.PELDS_NS + "matchedRule"), model
					.getIndividual(actionUri), ruleRes));

			try {
				Resource rule = swrl.rule(ruleUri, ruleModel
						.createList(antAtResList.iterator()), ruleModel
						.createList(conAtResList.iterator()));
				rule.addProperty(ruleModel.getProperty(RuleQuery.PELDS_NS
						+ "triplePatterns"), ruleModel.createList(patternList
						.iterator()));
				log.info("added Rule <" + ruleUri + ">:" + rule.toString());

			} catch (Exception e) {
				log.warning("failed to add rule <" + ruleUri + ">: "
						+ e.getMessage());
			}
		}
		ruleModel.setNsPrefixes(prefixMapping);
		return ruleModel;
	}

	/**
	 * Creates a SWRL rule Atom and adds it to a RDF model
	 * 
	 * @param ra
	 *            RuleAtom instance
	 * @param varMap
	 *            variables map
	 * @return RDF Resource representing the new atom
	 */
	private Resource addRuleAtom(RuleAtom ra, HashMap<String, Resource> varMap) {
		String uri = prefixMapping.expandPrefix(ra.getName());
		
		// sameAs atom
		if (uri.equals(SWRL.SameIndividualAtom.getURI())
				|| ra.getName().toLowerCase().endsWith("sameas")) {
			return swrl.sameAs(createParam(ra.getArguments().get(0), varMap),
					createParam(ra.getArguments().get(1), varMap));
		}

		// differentFrom atom
		if (uri.equals(SWRL.DifferentIndividualsAtom.getURI())
				|| ra.getName().toLowerCase().endsWith("differentfrom")) {
			return swrl.differentFrom(createParam(ra.getArguments().get(0),
					varMap), createParam(ra.getArguments().get(1), varMap));
		}

		// SWRL builtin atom
		if (uri.startsWith(Namespaces.SWRLB)) {
			List<RDFNode> paramList = new ArrayList<RDFNode>();
			for (String arg : ra.getArguments()) {
				paramList.add(createParamNode(arg,varMap,null));
			}
			return swrl.builtInAtom(ResourceFactory.createResource(uri),
					ruleModel.createList(paramList.iterator()));
		}

		// user-defined atoms
		Statement t = model.getProperty(model.getResource(uri), RDF.type);
		String typeUri;
		if (t != null) {
			typeUri = t.getObject().as(Resource.class).getURI();
		} else {
			typeUri = OWL.ObjectProperty.getURI();
			log.warning("Unable to find Property or Class <" + uri
					+ ">, falling back to ObjectProperty.");
		}
		// check if atom describes an class property
		if (typeUri.equals(OWL.Class.getURI()) || ra.getArguments().size() == 1) {
			Resource r = model.getResource(uri);
			if (r == null) {
				r = model.createClass(uri);
			}
			return swrl.classAtom(r, createParam(ra.getArguments().get(0),
					varMap));
		}
		// check if atom describes an object property
		if (typeUri.equals(OWL.ObjectProperty.getURI())) {
			ObjectProperty p = model.getObjectProperty(uri);
			if (p == null) {
				p = model.createObjectProperty(uri);
			}
			Resource arg0 = createParam(ra.getArguments().get(0), varMap);
			Resource arg1 = createParam(ra.getArguments().get(1), varMap);
			return swrl.individualPropertyAtom(p, arg0, arg1);
		}
		// check if atom describes an datatype property
		if (typeUri.equals(OWL.DatatypeProperty.getURI())) {
			DatatypeProperty p = model.getDatatypeProperty(uri);
			if (p == null) {
				p = model.createDatatypeProperty(uri);
			}
			Resource arg1 = createParam(ra.getArguments().get(0), varMap);
			RDFNode  arg2 = createParamNode(ra.getArguments().get(1), varMap, p);
			return swrl.dataPropertyAtom(p, arg1, arg2);
		}

		return null;
	}

	/**
	 * Creates a resource for every predicate/atom parameter
	 * 
	 * @param name
	 *            parameter name
	 * @param varMap
	 *            variables map containing already used parameters
	 * @return RDF resource representing variable
	 */
	private Resource createParam(String name, HashMap<String, Resource> varMap) {
		if (varMap.containsKey(name)) {
			return varMap.get(name);
		}
		// handle resources
		else {
			Resource r = model.createResource(prefixMapping.expandPrefix(name));
			model.add(model.createStatement(r, RDF.type, OWL.Thing));
			return r;
		}
	}

	/**
	 * Creates literal parameters for triple patterns, wraps a URL in a string if needed
	 * 
	 * @param name param value
	 * @param varMap variables map
	 * @return RDFNode containing a literal string
	 */
	private RDFNode createParamLiteral(String name,
			HashMap<String, Resource> varMap) {
		// handle literal values
		if (name.trim().startsWith("\"")) {
			Literal l = model.createTypedLiteral(name);
			return l;
		} else {
			return model.createTypedLiteral(createParam(name, varMap)
					.toString());
		}
	}
	
	/**
	 * Creates RDFNode parameters for atoms
	 * 
	 * @param name param value
	 * @param varMap variables map
	 * @return RDFNode containing a literal string or a resource node
	 */
	private RDFNode createParamNode(String name,
			HashMap<String, Resource> varMap, DatatypeProperty p) {
		// handle literal values
		if (name.trim().startsWith("\"")) {
			// find range
			name = name.replace("\"", "");
			Literal l;
			try {
				URI u = new URI(p.getRange().getURI());
				l = model.createTypedLiteral(name, new XSDDatatype(u.getFragment()));
			} catch (Exception e) {
				l = model.createLiteral(name);
			}
			return l;
		} else {
			return createParam(name,varMap);
		}
	}

	/**
	 * Strip all comments from rules
	 * 
	 * @param rules
	 *            rule string, with comments
	 * @return rule string, without comments
	 */
	private String removeComments(String rules) {
		String nr = rules.replaceAll(
				"(/\\*([^*]|[\\r\\n]|(\\*+([^*/]|[\\r\\n])))*\\*+/)", "");
		return nr;
	}

	public ArrayList<Rule> getRules() {
		return ruleList;
	}

	public void setVerbose(boolean set) {
		if (set) {
			log.setLevel(Level.ALL);
		} else {
			log.setLevel(Level.WARNING);
		}
	}
}
