package org.pelds.rules;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Collection;
import java.util.Iterator;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.mindswap.pellet.PelletOptions;
import org.mindswap.pellet.jena.PelletReasonerFactory;
import org.mindswap.pellet.utils.Namespaces;

import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.rdf.model.Selector;
import com.hp.hpl.jena.rdf.model.SimpleSelector;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.shared.PrefixMapping;
import com.hp.hpl.jena.update.UpdateAction;
import com.hp.hpl.jena.update.UpdateFactory;
import com.hp.hpl.jena.util.FileManager;
import com.hp.hpl.jena.util.FileUtils;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.RDF;

/**
 * RuleQuery takes a SPARQL query, a ruleset, and some data and returns only
 * triples permitted by the given rules.
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class RuleQuery {
	// Namespace and ontology
	public static final String PELDS_PREFIX = "pelds";
	public static final String PELDS_NS = "http://www.pelds.org/ont#";

	public static final String PELDS_ONT_LOC = "http://www.pelds.org/ont/pelds.owl";
	public static final String PELDS_ONT_FILE = "pelds.owl";

	// default input format
	private static final String DEFAULT_INPUT_FORMAT = "RDF/XML";

	// action labels
	public static final String QUERY_ACTION = "QueryAction";
	public static final String UPDATE_ACTION = "UpdateAction";

	// temporary working model, employs Pellet
	private OntModel model;

	// result model
	private Model checkedModel;

	// input model
	private Model data;

	// model to hold the PeLDS ontology
	private Model peldsOntModel;

	// collection of rules in their string representation
	private Collection<String> rules;

	// querying users Web IDs
	private Collection<String> identities;

	// query type, either QueryAction or UpdateAction
	private String queryType;

	// current actions URI
	private String actionUri;

	// base prefixes
	private static PrefixMapping prefixMapping = PrefixMapping.Factory.create()
			.setNsPrefixes(PrefixMapping.Extended).setNsPrefix(PELDS_PREFIX,
					PELDS_NS).setNsPrefix("swrl", Namespaces.SWRL).setNsPrefix(
					"swrlb", Namespaces.SWRLB);

	private Logger log = Logger.getLogger("RuleQuery");

	private boolean verbose = false;

	/**
	 * Constructor
	 * 
	 * @param type
	 *            Query type, either QueryAction or UpdateAction
	 * @param data
	 *            RDF Model containing the data to be checked
	 * @param identities
	 *            querying users authenticated Web IDs
	 * @param rules
	 *            rules in string representation
	 * @param pm
	 *            prefix mapping from configuration
	 */
	public RuleQuery(String type, Model data, Collection<String> identities,
			Collection<String> rules, PrefixMapping pm) {
		this.data = data;
		this.identities = identities;
		this.rules = rules;
		this.queryType = type;
		RuleQuery.prefixMapping.setNsPrefixes(pm);
	}

	/**
	 * Returns a model containing only the triples authorized by the rules
	 * 
	 * @return checked RDF model
	 */
	public Model getCheckedModel() {
		setVerbose(verbose);
		// configure model
		executeConfigure();
		// do reasoning
		executeReasoning();
		// filter authorized triples
		executeFilter();
		// results can now be fetched
		return checkedModel;
	}

	/**
	 * Returns true if the changes to the model are allowed by the rules
	 * 
	 * @param updates
	 *            changes as RDF model
	 * @return true if changes are allowed
	 */
	public boolean checkUpdates(Model updates) {
		setVerbose(verbose);
		// configure model
		executeConfigure();
		// do reasoning
		executeReasoning();
		// filter authorized triples
		return executeUpdateCheck(updates);
	}

	/**
	 * Initialize Pellet reasoner and working model, load PeLDS ontology
	 * 
	 * @throws RuleException
	 */
	private void executeConfigure() throws RuleException {
		// enable pellet rule support
		PelletOptions.DL_SAFE_RULES = true;
		// set strict rule handling as to not loose restrictions
		PelletOptions.IGNORE_UNSUPPORTED_AXIOMS = false;
		// disable classification monitor, craps up console
		PelletOptions.USE_CLASSIFICATION_MONITOR = PelletOptions.MonitorType.NONE;

		model = ModelFactory.createOntologyModel(
				PelletReasonerFactory.THE_SPEC, null);
		// use caching to speed up things around here
		// FileManager.get().setModelCaching(true);

		model.setNsPrefixes(prefixMapping);

		// add pelds ontology to model
		URL ontLocalPath = getClass().getResource(PELDS_ONT_FILE);
		peldsOntModel = addUriOrFile(ontLocalPath.toString());
		model.add(peldsOntModel);
	}

	/**
	 * Start the Pellet Reasoner
	 * 
	 * @throws RuleException
	 */
	private void executeReasoning() throws RuleException {
		// add data to model
		model.add(data);

		// add action object
		executeAddAction();
		// add rules to model
		addRules(rules);

		// prepare the model, check for consistency
		try {
			model.prepare();
		} catch (Exception e) {
			throw new RuleException("Unable to use rules", e);
		}
	}

	/**
	 * Add the action instance according to the PeLDS ontology
	 * 
	 */
	private void executeAddAction() {
		actionUri = createRandomizedUri("action");
		Resource action = model.createResource(actionUri);
		Property actor = model.createProperty(PELDS_NS + "actor");
		OntClass qac = model.getOntClass(PELDS_NS + queryType);

		model.add(action, RDF.type, qac);

		Iterator<String> idIt = identities.iterator();
		while (idIt.hasNext()) {
			String id = idIt.next();
			Resource user = model.createResource(id);
			model.add(action, actor, user);
		}
		return;
	}

	/**
	 * Create a random URI for variables and actions
	 * 
	 * @param subNs
	 *            namespace separator
	 * @return randomized URI
	 */
	public static String createRandomizedUri(String subNs) {
		// create a random uri for the action
		Random randomGenerator = new Random();
		int randomInt = randomGenerator.nextInt(Integer.MAX_VALUE);
		return PELDS_NS + subNs + "_" + String.valueOf(randomInt);
	}

	/**
	 * Execute the triple filtering process according to the triple patterns
	 * contained in the rules for this graph / model
	 * 
	 * @throws RuleException
	 */
	private void executeFilter() throws RuleException {
		// create a empty model on which the user query will be run
		checkedModel = ModelFactory.createDefaultModel();
		checkedModel.setNsPrefixes(prefixMapping);

		if (verbose)
			model.write(System.out, "TTL");

		// if no rules have matched exit
		if (!rulesMatched()) {
			log.info("No Rules matched");
			return;
		}
			
		// filter by triple / resource patterns
		ResultSet tRes = findTriplePatterns();
		while (tRes.hasNext()) {
			QuerySolution qs = tRes.next();
			String ruleStr = qs.get("?rule").toString();
			Selector selector = getSelector(qs);
			StmtIterator stmtIt = model.listStatements(selector);
			// add all permitted triples to the user query model
			checkedModel.add(stmtIt);
			if (verbose)
				log.info("Rule " + ruleStr + " approved triples ("
						+ selector.getSubject() + ", "
						+ selector.getPredicate() + ", " + selector.getObject()
						+ ")");
		}
		
		// filter by instance patterns
		ResultSet iRes = findInstancePatterns();
		while (iRes.hasNext()) {
			QuerySolution qs = iRes.next();
			String ruleStr = qs.get("?rule").toString();

			String instanceStr = qs.getLiteral("?instance").getString()
					.replace("*", "");

			// find all individuals with that type, approve all statements
			// describing them
			ExtendedIterator<Individual> instanceIt = model
					.listIndividuals(ResourceFactory
							.createResource(instanceStr));
			while (instanceIt.hasNext()) {
				Individual ind = instanceIt.next();
				checkedModel.add(ind.listProperties());
			}

			if (verbose)
				log.info("Rule " + ruleStr + " approved instances of class <"
						+ instanceStr + ">");
		}
		
		// remove PeLDS ontology & its instances again
		// otherwise rules are published, which is bad.
		// note: some rdf:list clutter is left, but should not matter
		removePeldsInstances(checkedModel);

		ResIterator resIt = peldsOntModel.listSubjects();
		while (resIt.hasNext()) {
			Resource peldsOntRes = resIt.next();
			checkedModel.removeAll(peldsOntRes, null, null);
			checkedModel.removeAll(null, null, peldsOntRes);
		}

		if (verbose)
			checkedModel.write(System.out, "TTL");
	}

	/**
	 * Removes all instances of the PeLDS vocabulary from the model
	 * 
	 * @param m model to be updated
	 */
	private void removePeldsInstances(Model m) {
		String patternDelQuery = "PREFIX pelds: <"
				+ PELDS_NS
				+ "> \n"
				+ "PREFIX list: <http://jena.hpl.hp.com/ARQ/list#> \n"
				+ "PREFIX swrl: <http://www.w3.org/2003/11/swrl#> \n"

				// delete triple patterns
				+ "DELETE { ?s ?p ?o} \n"
				+ "WHERE { \n"
				+ " ?s ?p ?o . \n"
				+ " ?rule 	a pelds:Rule ; \n"
				+ " 		pelds:triplePatterns ?patternList.\n"
				+ " ?patternList list:member ?s. \n"
				+ "} \n"

				// delete rule atoms
				+ "DELETE { ?s ?p ?o} \n" + "WHERE { \n" + " ?s ?p ?o . \n"
				+ " ?rule 	a pelds:Rule ; \n" + " 		?someList ?atomList.\n"
				+ " ?atomList list:member ?s. \n"
				+ "} \n"

				// delete rule instances
				+ "DELETE { ?s ?p ?o} \n" + "WHERE { \n" + " ?s ?p ?o . \n"
				+ " ?s 	a pelds:Rule . \n" + "} \n"

				// delete action instance
				+ "DELETE { ?s ?p ?o} \n" + "WHERE { \n" + " ?s ?p ?o . \n"
				+ " ?s 	a pelds:Action . \n" + "} \n"

				+ "\n";

		UpdateAction.execute(UpdateFactory.create(patternDelQuery),
				checkedModel);
	}

	/**
	 * Check if given updates are permitted by the rules for this graph / model
	 * 
	 * @param updatesModel
	 *            the requested updates
	 * @return true if the changes are permitted
	 */
	private boolean executeUpdateCheck(Model updatesModel) {
		// if no rules have matched exit
		if (!rulesMatched()) {
			log.info("No Rules matched");
			return false;
		}

		ResultSet tRes = findTriplePatterns();
		while (tRes.hasNext()) {
			// shortcut if we should be already done
			if (updatesModel.size() == 0)
				return true;

			QuerySolution qs = tRes.next();
			String ruleStr = qs.get("?rule").toString();
			Selector selector = getSelector(qs);
			// remove all permitted triples from the updateGraph
			updatesModel.remove(updatesModel.listStatements(selector));
			if (verbose)
				log.info("Rule " + ruleStr + " approved triples ("
						+ selector.getSubject() + ", "
						+ selector.getPredicate() + ", " + selector.getObject()
						+ ")");
		}

		// shortcut if we should be already done
		if (updatesModel.size() == 0)
			return true;

		ResultSet iRes = findInstancePatterns();
		while (iRes.hasNext()) {
			// shortcut if we should be already done
			if (updatesModel.size() == 0)
				return true;

			QuerySolution qs = iRes.next();
			String ruleStr = qs.get("?rule").toString();
			// TODO: TEST THIS!
			String instanceStr = qs.getLiteral("?instance").getString()
					.replace("*", "");

			// find all individuals with that type, approve all statements
			// describing them
			ExtendedIterator<Individual> instanceIt = model
					.listIndividuals(ResourceFactory
							.createResource(instanceStr));
			while (instanceIt.hasNext()) {
				Individual ind = instanceIt.next();
				updatesModel.remove(ind.listProperties());
			}
			if (verbose)
				log.info("Rule " + ruleStr + " approved instances of class <"
						+ instanceStr + ">");
		}
		// return true if no statements are left, if that is the case, all
		// updates have been approved
		boolean res = updatesModel.size() == 0;
		if (verbose && !res) {
			System.out.println("Unauthorized Triples:");
			updatesModel.write(System.out, "TTL");
		}
		return res;
	}

	/**
	 * Create a triple selector from the triple pattern
	 * 
	 * @param qs
	 *            SPARQL result containing the pattern
	 * @return triple selector
	 */
	private Selector getSelector(QuerySolution qs) {
		// convert triple pattern to SimpleSelector
		Resource subjNode = null;
		Property predNode = null;
		RDFNode objeNode = null;

		String subjStr = qs.getLiteral("?subject").getString().replace("*", "");
		String predStr = qs.getLiteral("?predicate").getString().replace("*",
				"");
		String objeStr = qs.getLiteral("?object").getString().replace("*", "");

		if (!subjStr.equals(""))
			subjNode = ResourceFactory.createResource(subjStr);
		if (!predStr.equals(""))
			predNode = ResourceFactory.createProperty(predStr);
		if (!objeStr.equals(""))
			objeNode = ResourceFactory.createPlainLiteral(objeStr);

		return new SimpleSelector(subjNode, predNode, objeNode);
	}

	/**
	 * Find triple patterns to be authorized
	 * 
	 * @return SPARQL result set containing patterns
	 */
	private ResultSet findTriplePatterns() {
		// find triple patterns for the matched rules with permitted status
		String patternQueryString = "PREFIX pelds: <" + PELDS_NS + "> \n"
				+ "PREFIX list: <http://jena.hpl.hp.com/ARQ/list#> \n"
				+ "SELECT ?rule ?subject ?predicate ?object \n" + "WHERE { \n"
				+ "      <" + actionUri + "> pelds:matchedRule ?rule. \n"
				+ "      ?rule pelds:triplePatterns ?patternList.\n"
				+ "  	   ?patternList list:member ?pattern. \n"
				+ "      ?pattern a pelds:TriplePattern ; \n"
				+ "      		pelds:tripleSubject ?subject ; \n"
				+ "      		pelds:triplePredicate ?predicate ; \n"
				+ "      		pelds:tripleObject ?object . \n" + "} \n";

		ResultSet tRes = execSparql(patternQueryString);
		return tRes;
	}

	/**
	 * Find instance patterns to be authorized
	 * 
	 * @return SPARQL result set containg patterns
	 */
	private ResultSet findInstancePatterns() {
		// find instance patterns for the matched rules with permitted status
		String instanceQueryString = "PREFIX pelds: <" + PELDS_NS + "> \n"
				+ "PREFIX list: <http://jena.hpl.hp.com/ARQ/list#> \n"
				+ "SELECT ?rule ?instance \n" + "WHERE { \n" + "      <"
				+ actionUri + "> pelds:matchedRule ?rule. \n"
				+ "      ?rule pelds:triplePatterns ?patternList.\n"
				+ "  	   ?patternList list:member ?pattern. \n"
				+ "      ?pattern a pelds:TriplePattern ; \n"
				+ "      		pelds:instance ?instance . \n" + " } \n";

		ResultSet iRes = execSparql(instanceQueryString);
		return iRes;
	}

	/**
	 * Check wether any rule for this graph / model has matched
	 * 
	 * @return true if any rule matched
	 */
	private boolean rulesMatched() {
		String matchCountQueryString = "PREFIX pelds: <" + PELDS_NS + "> \n"
				+ "SELECT ?rule \n" + "WHERE { \n" + "      <" + actionUri
				+ "> pelds:matchedRule ?rule. \n" + "} \n";

		ResultSet countRes = execSparql(matchCountQueryString);
		int count = 0;
		while (countRes.hasNext()) {
			QuerySolution qs = countRes.next();
			log.info("Matched Rule: " + qs.get("?rule").toString());
			count++;
		}
		return count > 0;
	}

	/**
	 * Add a graph to the model, can be a remote URL
	 * 
	 * @param urifile
	 *            local file name or remote URL
	 */
	private Model addUriOrFile(String urifile) {
		if (verbose)
			log.info("Adding data File/URI <" + urifile + ">");
		Model m = ModelFactory.createDefaultModel();
		String lang = FileUtils.guessLang(urifile,
				RuleQuery.DEFAULT_INPUT_FORMAT);
		try {
			FileManager.get().readModel(m, urifile, lang);
		} catch (Exception e) {
			log.warning("Failed to open data File/URI <" + urifile + ">");
		}
		return m;
	}

	/**
	 * Execute a SPARQL query on the working model
	 * 
	 * @param queryStr
	 *            the SPARQL query
	 * @return the result set
	 */
	private ResultSet execSparql(String queryStr) {
		return execSparql(queryStr, model);
	}

	/**
	 * Execute a SPARQL query on a model
	 * 
	 * @param queryStr
	 *            the SPARQL query
	 * @param m
	 *            the model
	 * @return the result set
	 */
	private ResultSet execSparql(String queryStr, Model m) {
		if (queryStr == null) {
			log.warning("No SPARQL Query found!");
		}
		Query query = QueryFactory.create(queryStr);
		// UpdateRequest u = UpdateFactory.create(queryStr);

		QueryExecution qe = QueryExecutionFactory.create(query, m);
		return qe.execSelect();
	}

	/**
	 * Parse and add rules to the working model
	 * 
	 * @param rules
	 *            rules in text representation
	 */
	private void addRules(Collection<String> rules) {
		Iterator<String> ruleIt = rules.iterator();
		while (ruleIt.hasNext()) {
			String rule = ruleIt.next();
			try {
				RuleParser rp = new RuleParser(rule, verbose);
				Model ruleModel = rp.addRulesToModel(getModel(),
						getPrefixMapping(), actionUri);

				if (verbose)
					ruleModel.write(System.out, "TTL");
				if (verbose)
					log.info(rp.getRules().toString());
				model.add(ruleModel);

			} catch (Exception e) {
				log.warning("Failed to add rule " + rule);
				e.printStackTrace();
			}
		}
	}

	/*
	 * if (verbose) model.write(System.out,"TTL"); }
	 * 
	 * // TODO: use below code to allow prettier patterns?
	 * 
	 * private static Triple parseTriplePattern(String s) { String[] nodes =
	 * s.split(" +", 3); if (nodes.length != 3) { throw new
	 * IllegalArgumentException("-find: Failed to parse triple pattern"); } Node
	 * subject = parseNode(nodes[0], PrefixMapping.Extended); Node predicate =
	 * parseNode(nodes[1], PrefixMapping.Extended); Node object =
	 * parseNode(nodes[2], PrefixMapping.Extended); return new Triple(subject,
	 * predicate, object); }
	 * 
	 * private static Node parseNode(String s, PrefixMapping prefixes) { if
	 * ("ANY".equals(s)) { return Node.ANY; } Matcher m =
	 * Pattern.compile("^_:(.*)$").matcher(s); if (m.matches()) { return
	 * Node.createAnon(new AnonId(m.group(1))); } m =
	 * Pattern.compile("^<(.*)>$").matcher(s); if (m.matches()) { return
	 * Node.createURI(m.group(1)); } m =
	 * Pattern.compile("^([a-zA-Z0-9_-]*):(.*)$").matcher(s); if (m.matches()) {
	 * if (prefixes.getNsPrefixURI(m.group(1)) == null) { throw new
	 * IllegalArgumentException( "-find: Unknown prefix: '" + m.group(1) + "'");
	 * } return Node.createURI(prefixes.getNsPrefixURI(m.group(1)) +
	 * m.group(2)); } m = Pattern.compile("^(?:'.*'|\".*\")$").matcher(s); if
	 * (m.matches()) { return Node.createLiteral(m.group(1).substring(1,
	 * m.group(1).length() - 2)); } m =
	 * Pattern.compile("^(?:'.*'|\".*\")@([a-zA-Z0-9_-]+)$").matcher(s); if
	 * (m.matches()) { return Node.createLiteral(m.group(1).substring(1,
	 * m.group(1).length() - 2), m.group(2), null); } m =
	 * Pattern.compile("^(?:'.*'|\".*\")^^(.*)$").matcher(s); if (m.matches()) {
	 * return Node.createLiteral(m.group(1).substring(1, m.group(1).length() -
	 * 2), null, TypeMapper.getInstance().getSafeTypeByName(
	 * parseNode(m.group(2), prefixes).getURI())); } throw new
	 * IllegalArgumentException("Failed to parse node: '" + s + "'"); }
	 */

	/**
	 * Reads a whole InputStream into a String
	 * 
	 * @param in
	 *            Stream
	 * @return String with the data from the stream
	 * @throws IOException
	 */
	public static String slurp(InputStream in) throws IOException {
		StringBuffer out = new StringBuffer();
		byte[] b = new byte[4096];
		for (int n; (n = in.read(b)) != -1;) {
			out.append(new String(b, 0, n));
		}
		return out.toString();
	}

	public RuleQuery setIdentities(Collection<String> verifiedIds) {
		this.identities = verifiedIds;
		return this;
	}

	public RuleQuery setVerbose(boolean set) {
		if (set) {
			log.setLevel(Level.ALL);
			verbose = true;
		} else {
			log.setLevel(Level.INFO);
			verbose = false;
		}
		return this;
	}

	public RuleQuery setRules(Collection<String> rules) {
		this.rules = rules;
		return this;
	}

	public RuleQuery addRule(String ruleStr) {
		rules.add(ruleStr);
		return this;
	}

	public RuleQuery setData(Model data) {
		this.data = data;
		return this;
	}

	public PrefixMapping getPrefixMapping() {
		return prefixMapping;
	}

	public OntModel getModel() {
		return model;
	}

}
