package org.pelds.servlet;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Collection;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.pelds.store.NamedGraphStore;

/**
 * Handles Linked Data derferencing requests, creates a SPARQL DESCRIBE QUERY by
 * looking up the named graphs the requested resource appears in, and forwards
 * to an URL containing this query
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class DereferencingRequestServlet extends Servlet {

	private static final long serialVersionUID = 1L;
	static private Logger log = Logger.getLogger("DereferencingRequestServlet");

	/**
	 * @see org.pelds.servlet.Servlet#service(javax.servlet.http.HttpServletRequest,
	 *      javax.servlet.http.HttpServletResponse)
	 */
	protected void service(HttpServletRequest req, HttpServletResponse resp) {
		// no parameter class, as parameters are not needed...

		String reqUri = req.getRequestURI(); // /pelds/data/horststein
		String context = req.getContextPath() + req.getServletPath(); // /pelds/data
		reqUri = reqUri.replace(context, ""); // /horststein
		reqUri = getConfig().getDereferencingPrefix() + reqUri; // http://example.com/horststein
		log.info("Start processing request for resource <" + reqUri + ">.");

		// 1 find datasets for reqUri
		NamedGraphStore ngs = new NamedGraphStore();
		ngs.setConfig(getConfig());

		Collection<String> datasets = ngs.findDatasetForResource(reqUri);

		// if we cannot find a dataset, send an error
		if (datasets.size() == 0) {
			try {
				resp.sendError(HttpServletResponse.SC_NOT_FOUND);
			} catch (IOException e) {
				// do nothing
			}
			return;
		}

		// 2 construct DESCRIBE & parameters
		String query = "";
		try {
			query = "?query="
					+ URLEncoder.encode("DESCRIBE <" + reqUri + ">", "UTF-8");
			for (String dataset : datasets) {
				query = query + "&default-graph-uri="
						+ URLEncoder.encode(dataset, "UTF-8");
			}
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}

		// 3 redirect to query
		String redirect = getConfig().getDereferencingEndpoint() + query;

		resp.setStatus(HttpServletResponse.SC_MOVED_TEMPORARILY);
		resp.setHeader("Location", redirect);

		log.info("Finished processing request by redirecting to " + redirect);

	}
}