package org.pelds.servlet;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import javax.servlet.http.HttpServletRequest;

import org.pelds.Constants;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryException;
import com.hp.hpl.jena.query.QueryFactory;

/**
 * Prepares the request parameters for the query request servlet
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 * @see QueryRequestServlet
 */
public class QueryRequestParameters implements RequestParameters {
	// parameters
	protected String queryString;
	protected Query query;
	protected String responseContentType;
	protected Collection<String> datasetUris;

	// errors
	protected String errorMsgs;

	// default response content type
	static final public String dfltResponseContentType = Constants.MIME_TYPE_RESULT_XML;

	/**
	 * @see org.pelds.servlet.RequestParameters#process(javax.servlet.http.HttpServletRequest)
	 */
	public boolean process(HttpServletRequest req) {
		errorMsgs = "";

		// force utf-8
		try {
			req.setCharacterEncoding("UTF-8");
		} catch (UnsupportedEncodingException e1) {
			// we'll just assume utf is well-known by 2009
		}

		// - the SPARQL query
		query = null;
		queryString = req.getParameter("query");
		if (queryString == null || queryString.equals("")) {
			errorMsgs += "Query not specified. ";
		} else {
			try {
				query = QueryFactory.create(queryString);
				if (!query.isSelectType() && !query.isDescribeType()
						&& !query.isConstructType()) {
					errorMsgs += "Query must be a SELECT or DESCRIBE query. ";
				}
				datasetUris = new ArrayList<String>();
				if (query.getGraphURIs().size() == 0) {
					String[] dataSetParams = req
							.getParameterValues("default-graph-uri");
					if (dataSetParams != null) {
						datasetUris.addAll(Arrays.asList(dataSetParams));
					}
					if (datasetUris.size() == 0) {
						errorMsgs += "Please specify at least one dataset (SELECT FROM <...>) in your query or with a GET parameter. ";
					}
				} else {
					datasetUris = query.getGraphURIs();
				}
			} catch (QueryException e) {
				errorMsgs += e.getMessage() + " ";
			}
		}

		// determine the output type...
		responseContentType = null;
		// either by GET parameter
		String output = req.getParameter("output");
		if (output != null) {
			if (output.equals("json")) {
				responseContentType = Constants.MIME_TYPE_RESULT_JSON;
			} else if (output.equals("xml")) {
				responseContentType = Constants.MIME_TYPE_RESULT_XML;
			} else if (output.equals("txt")) {
				responseContentType = Constants.MIME_TYPE_TEXT;
			} else {
				errorMsgs += "Unsupported output format requested. ";
			}
		}
		// or by HTTP header
		else {
			String accept = req.getHeader("ACCEPT");
			if (accept != null
					&& (accept.contains(Constants.MIME_TYPE_RESULT_JSON) || accept
							.contains(Constants.MIME_TYPE_JSON))) {
				responseContentType = Constants.MIME_TYPE_RESULT_JSON;
			} else if (accept != null
					&& (accept.contains(Constants.MIME_TYPE_RESULT_XML)
							|| accept.contains(Constants.MIME_TYPE_XML1) || accept
							.contains(Constants.MIME_TYPE_XML2))) {
				responseContentType = Constants.MIME_TYPE_RESULT_XML;
			} else {
				responseContentType = dfltResponseContentType;
			}
		}

		return (errorMsgs.equals(""));
	}

	public String getErrorMsgs() {
		return errorMsgs;
	}

	public String getQueryString() {
		return queryString;
	}

	public Query getQuery() {
		return query;
	}

	public String getResponseContentType() {
		return responseContentType;
	}

	public Collection<String> getDatasetUris() {
		return datasetUris;
	}

}