package org.pelds.servlet;

import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.ResultSetFormatter;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.sparql.resultset.ResultSetFormat;
import com.hp.hpl.jena.sparql.resultset.ResultSetMem;
import com.hp.hpl.jena.sparql.util.Utils;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.mindswap.pellet.PelletOptions;
import org.mindswap.pellet.jena.PelletReasonerFactory;
import org.pelds.Constants;
import org.pelds.auth.FoafSslInterface;
import org.pelds.rules.RuleQuery;
import org.pelds.store.DatasetOwnerStore;
import org.pelds.store.NamedGraphStore;
import org.pelds.store.RuleStore;
import org.pelds.util.Timer;

/**
 * Processes a SPARQL SELECT, DESCRIBE or CONSTRUCT query, evaluates the access
 * policy for the RDF Graph mentioned in the query, loads the graph, determines
 * the authorized triples, executes the query on the authorized triples, and
 * formats the results.
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class QueryRequestServlet extends Servlet {

	private static final long serialVersionUID = 1L;
	static private Logger log = Logger.getLogger("QueryRequestServlet");

	/**
	 * @see org.pelds.servlet.Servlet#service(javax.servlet.http.HttpServletRequest,
	 *      javax.servlet.http.HttpServletResponse)
	 */
	protected void service(HttpServletRequest req, HttpServletResponse resp) {
		Timer overallTimer = new Timer("overall").start();

		// content negotiation
		String accept = req.getHeader("ACCEPT");
		if (accept != null && !accept.contains(Constants.MIME_TYPE_RESULT_XML)
				&& !accept.contains(Constants.MIME_TYPE_XML1)
				&& !accept.contains(Constants.MIME_TYPE_XML2)
				&& !accept.contains(Constants.MIME_TYPE_RESULT_JSON)
				&& !accept.contains(Constants.MIME_TYPE_JSON)
				&& !accept.contains(Constants.MIME_TYPE_RESULT_RDF)
				&& !accept.contains(Constants.MIME_TYPE_RESULT_N3)

				&& !accept.contains("application/*") && !accept.contains("*/*")) {

			sendError(req, resp,
					"Your client does not seem to accept one of the possible content types (e.g. '"
							+ Constants.MIME_TYPE_RESULT_XML + "', '"
							+ Constants.MIME_TYPE_RESULT_RDF + "', '"
							+ Constants.MIME_TYPE_RESULT_N3 + "', '"
							+ Constants.MIME_TYPE_RESULT_JSON + "').");
			return;
		}

		// set some headers to let people know what is going on here
		resp.addHeader("Server", "PeLDS");
		resp.addHeader("X-Powered-By", "Policy enabled Linked Data Server (PeLDS), see http://www.pelds.org for details.");
		resp.addHeader("X-Note", "The query result you are recieving may be restricted by access policies. Consider using FOAF+SSL authentication to get full results.");

		// get (and check) the request parameters
		QueryRequestParameters params = new QueryRequestParameters();
		if (!params.process(req)) {
			sendError(req, resp, params.getErrorMsgs());
			return;
		}

		// some logging
		log.info("Start executing request with query:");
		log.info(params.getQueryString());

		Timer authTimer = new Timer("auth").start();
		// do FOAF + SSL authentication
		Collection<String> verifiedIds = FoafSslInterface
				.getVerifiedWebIds(req);
		log.info(" Authenticated IDs: " + verifiedIds.toString());
		authTimer.stop();

		NamedGraphStore ngs = new NamedGraphStore();
		ngs.setConfig(getConfig());

		Model queryModel = ModelFactory.createDefaultModel();

		// a sparql query can contain multiple dataset references, so handle
		// them all
		for (String datasetUri : params.getDatasetUris()) {
			Timer loadDataTimer = new Timer("load").start();
			// load requested graph from store
			Model dbModel = ngs.loadModel(datasetUri);

			DatasetOwnerStore dos = new DatasetOwnerStore(datasetUri);

			// check owner, grant full access if user is owner
			if (dos.isOwner(datasetUri, verifiedIds)) {

				PelletOptions.USE_CLASSIFICATION_MONITOR = PelletOptions.MonitorType.NONE;
				OntModel reasoningModel = ModelFactory.createOntologyModel(
						PelletReasonerFactory.THE_SPEC, null);
				reasoningModel.add(dbModel);
				reasoningModel.prepare();
				queryModel.add(reasoningModel);

				log.info("Owner access to <" + datasetUri + ">");
				continue;
			}

			log.info("Evaluating rules for dataset <" + datasetUri + ">");
			// load rules (HR format)
			Collection<String> ruleList = new ArrayList<String>();
			ruleList.add(new RuleStore(datasetUri).getRuleStr());
			loadDataTimer.stop();

			Timer ruleTimer = new Timer("rules").start();
			// create rule checking instance
			Model checkedModel = ModelFactory.createDefaultModel();
			try {
				checkedModel = new RuleQuery(RuleQuery.QUERY_ACTION, dbModel,
						verifiedIds, ruleList, getConfig().getPrefixMap())
						.getCheckedModel();
			} catch (Exception e2) {
				log.warning("Evaluating rules for " + req.hashCode()
						+ " caused a " + Utils.className(e2) + ": "
						+ e2.getMessage());
				e2.printStackTrace();
			}
			ruleTimer.stop();
			queryModel.add(checkedModel);
		}

		// set response encoding, just to be sure
		resp.setCharacterEncoding("UTF-8");

		Timer sparqlTimer = new Timer("sparql").start();
		// execute SPARQL
		ResultSetMem resultSet = null;
		QueryExecution qe = QueryExecutionFactory.create(params.getQuery(),
				queryModel);

		// DESCRIBE Query
		if (params.getQuery().isDescribeType()) {
			Model descrResModel = qe.execDescribe();
			try {
				resp.setContentType(Constants.MIME_TYPE_RESULT_RDF);
				resp.getOutputStream().write(
						"<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"
								.getBytes());
				descrResModel.write(resp.getOutputStream(), "RDF/XML-ABBREV");
			} catch (IOException e) {
				sendError(req, resp, e.getMessage());
			}
			return;
		}

		// SELECT Query
		if (params.getQuery().isSelectType()) {
			resultSet = new ResultSetMem(qe.execSelect());
			sparqlTimer.stop();

			// create the response
			OutputStream out = null;
			try {
				out = resp.getOutputStream();
			} catch (IOException e) {
				sendError(req, resp, "I/O Error");
				return;
			}

			// write the response
			resp.setContentType(params.getResponseContentType());
			try {
				if (params.getResponseContentType() == Constants.MIME_TYPE_RESULT_JSON) {
					ResultSetFormatter.outputAsJSON(out, resultSet);
				} else if (params.getResponseContentType() == Constants.MIME_TYPE_TEXT) {
					ResultSetFormatter.output(out, resultSet,
							ResultSetFormat.syntaxText);
				} else {
					ResultSetFormatter.outputAsXML(out, resultSet);
				}
			} catch (Exception e) {
				sendError(req, resp, "I/O Error");
			}
		}

		// close persistence
		try {
			ngs.closeGraphSet();
			ngs.closeConnection();
		} catch (Exception e) {
			log.warning("Failed to close persistence layer");
		}
		// finish
		try {
			resp.getOutputStream().flush();
			resp.flushBuffer();
			resp.getOutputStream().close();
		} catch (IOException e) {
			log.warning("Flushing the response buffer for request "
					+ req.hashCode() + " caused a " + Utils.className(e) + ": "
					+ e.getMessage());
		}
		overallTimer.stop();

	}
}