package org.pelds.servlet;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.pelds.Constants;
import org.pelds.auth.FoafSslInterface;
import org.pelds.rules.Rule;
import org.pelds.rules.RuleParser;
import org.pelds.store.DatasetOwnerStore;
import org.pelds.store.NamedGraphStore;
import org.pelds.store.RuleStore;

import com.hp.hpl.jena.sparql.util.Utils;

/**
 * Updates the access policy for a given RDF dataset. Policy has to be given in
 * PsSF format, user can only set / change the policy if she created the
 * corresponding graph, if no graph with the given identifier exists, it is
 * created.
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class RuleRequestServlet extends Servlet {

	private static final long serialVersionUID = 1L;
	static private Logger log = Logger.getLogger("RuleRequestServlet");

	/**
	 * @see org.pelds.servlet.Servlet#service(javax.servlet.http.HttpServletRequest,
	 *      javax.servlet.http.HttpServletResponse)
	 */
	protected void service(HttpServletRequest req, HttpServletResponse resp) {
		log.info("Start processing request " + req.hashCode() + ".");

		// get (and check) the request parameters
		RuleRequestParameters params = new RuleRequestParameters();
		if (!params.process(req)) {
			sendError(req, resp, params.getErrorMsgs());
			return;
		}

		// some logging
		log.info("Start rule update for <" + params.getDatasetUri()
				+ "> with rules:");
		log.info(params.getRuleStr());

		// do FOAF + SSL authentication
		Collection<String> verifiedIds = FoafSslInterface
				.getVerifiedWebIds(req);
		log.info(" Authenticated IDs: " + verifiedIds.toString());

		// check if user has rights to dataset, if no ds exists, create one
		NamedGraphStore ngs = new NamedGraphStore();
		ngs.setConfig(getConfig());
		String datasetUri = params.getDatasetUri();
		ngs.loadOrCreateModel(datasetUri, verifiedIds);
		DatasetOwnerStore dos = new DatasetOwnerStore(datasetUri);

		if (!dos.isOwner(datasetUri, verifiedIds)) {
			try {
				resp
						.sendError(HttpServletResponse.SC_FORBIDDEN,
								"Only the dataset creator may see/change the access policy");
			} catch (IOException e) {
			}
			return;
		}

		RuleStore rs = new RuleStore(params.getDatasetUri());
		rs.setConfig(getConfig());

		String ruleStr = "";
		if (params.hasRuleString()) {
			ruleStr = params.getRuleStr();
			log.info("Using rules from Query");
		} else {
			ruleStr = rs.getRuleStr();
			log.info("Using rules from DB");
		}

		List<Rule> ruleList = new ArrayList<Rule>();
		// try to parse rules
		try {
			ruleList = new RuleParser(ruleStr, false).getRules();
		} catch (Exception e) {
			log
					.warning("Failed to parse rules: " + req.hashCode()
							+ " caused a " + Utils.className(e) + ": "
							+ e.getMessage());
			sendError(req, resp, "Failed to parse rules");
			return;
		}

		// save ruleset (if given, otherwise print existing rules)
		if (params.hasRuleString()) {
			log.info("Saving rules for dataset <" + params.getDatasetUri()
					+ "> :" + params.getRuleStr());
			rs.setRuleStr(params.getRuleStr());
		}

		// print ruleset
		String outStr = "";
		if (ruleList.size() > 0) {
			outStr = "/* Rule Compilation Result */\n\n";
			for (Rule r : ruleList) {
				outStr += r.toString() + "\n";
			}
			outStr += "/* EOF */";
		} else {
			sendError(req, resp, "No rules found nor set for dataset <"
					+ params.getDatasetUri() + ">");
			return;
		}
		// write the response
		resp.setContentType(Constants.MIME_TYPE_TEXT);
		try {
			resp.getWriter().write(outStr);
		} catch (Exception e) {
			log.warning("Writing to the response stream for request "
					+ req.hashCode() + " caused a " + Utils.className(e) + ": "
					+ e.getMessage());
		}

	}
}