package org.pelds.servlet;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.Logger;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.pelds.Config;

import com.hp.hpl.jena.sparql.util.Utils;

/**
 * Base class for all PeLDS servlets, provides access to configuration and
 * supports error communication
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
abstract public class Servlet extends HttpServlet {
	private static final long serialVersionUID = 1L;
	static private Logger log = Logger.getLogger("Servlet");
	static private Context namingContext = null;

	/**
	 * Retrieve PeLDS configuration instance
	 * 
	 * @return Config instance
	 * @see org.pelds.Config
	 */
	final protected Config getConfig() {
		Config config = (Config) getServletContext().getAttribute(
				"org.pelds.servlet.config");
		if (config == null) {
			config = new Config();

			Properties p = new Properties();
			// NOTE: configuration file location is set in web.xml

			String cf = getConfigFileLocation();
			try {
				File fi = new File(cf);
				p.load(new FileInputStream(fi));
				config.setConfigDir(fi.getParent());
				log.info("Using config file " + fi.getAbsolutePath());
			} catch (FileNotFoundException e) {
				log.warning("Failed to find config file '" + cf
						+ "' - using defaults");
			} catch (IOException e) {
				log.warning("Failed to read config file '" + cf + "': "
						+ e.getMessage());
			}
			config.init(p);
			getServletContext()
					.setAttribute("org.pelds.servlet.config", config);
		}
		return config;
	}

	/**
	 * Sends and logs a BAD REQUEST error
	 * 
	 * @param req
	 *            Request object
	 * @param resp
	 *            Response object
	 * @param msg
	 *            Error message
	 */
	final protected void sendError(HttpServletRequest req,
			HttpServletResponse resp, String msg) {
		try {
			log.info("BAD REQUEST for request " + req.hashCode() + ": " + msg);
			resp.sendError(HttpServletResponse.SC_BAD_REQUEST, msg);
		} catch (IOException e) {
			log.warning("Sending the error reponse caused a "
					+ Utils.className(e) + ": " + e.getMessage());
		}
	}

	/**
	 * Configuration property file location is read from web.xml
	 * 
	 * @return Location of configuration file
	 */
	static protected String getConfigFileLocation() {
		return getStringFromContext("ConfigFileLocation");
	}

	/**
	 * Get String from JDNI context
	 * 
	 * @param key
	 *            context value key
	 * @return context value
	 */
	static private String getStringFromContext(String key) {
		Context nCtx = getNamingContext();
		try {
			return (String) nCtx.lookup(key);
		} catch (NameNotFoundException e) {
			return null;
		} catch (NamingException e) {
			log.warning("Looking up the object named '" + key + "' caused a "
					+ Utils.className(e) + ": " + e.getMessage());
			return null;
		}
	}

	/**
	 * Retrieve JDNI context
	 * 
	 * @return JDNI context
	 */
	static protected Context getNamingContext() {
		if (namingContext == null) {
			try {
				namingContext = (Context) (new InitialContext())
						.lookup("java:comp/env");
			} catch (NamingException e) {
				throw new Error(
						"Looking up the application naming context caused a "
								+ Utils.className(e) + ": " + e.getMessage(), e);
			}
		}

		return namingContext;
	}

	/**
	 * Handles HTTP GET and POST requests
	 * 
	 * @param req
	 *            HTTP Request Object
	 * @param resp
	 *            HTTP Response Object
	 * 
	 * @see javax.servlet.http.HttpServlet#service(javax.servlet.http.HttpServletRequest,
	 *      javax.servlet.http.HttpServletResponse)
	 */
	abstract protected void service(HttpServletRequest req,
			HttpServletResponse resp);
}