package org.pelds.servlet;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.pelds.auth.FoafSslInterface;
import org.pelds.rules.RuleQuery;
import org.pelds.rules.UpdateGraphListener;
import org.pelds.store.DatasetOwnerStore;
import org.pelds.store.NamedGraphStore;
import org.pelds.store.RuleStore;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.sparql.util.Utils;
import com.hp.hpl.jena.update.GraphStore;
import com.hp.hpl.jena.update.GraphStoreFactory;
import com.hp.hpl.jena.update.UpdateAction;

/**
 * Processes a SPARQL/Update request on one given dataset, evaluates the access
 * policy for the RDF Graph mentioned in the query, loads the graph, executes
 * the query and checks if the changes were permitted by the given access
 * policy. If so, the graph is saved.
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class UpdateRequestServlet extends Servlet {

	private static final long serialVersionUID = 1L;
	static private Logger log = Logger.getLogger("UpdateRequestServlet");

	/**
	 * @see org.pelds.servlet.Servlet#service(javax.servlet.http.HttpServletRequest,
	 *      javax.servlet.http.HttpServletResponse)
	 */
	protected void service(HttpServletRequest req, HttpServletResponse resp) {
		// get (and check) the request parameters
		UpdateRequestParameters params = new UpdateRequestParameters();
		if (!params.process(req)) {
			log.info("BAD REQUEST: " + params.getErrorMsgs());
			sendError(req, resp, params.getErrorMsgs());
			return;
		}

		// some logging
		log.info("Start executing update on <" + params.getDatasetUris()
				+ "> with update:");
		log.info(params.getQueryString());

		// do FOAF + SSL authentication
		Collection<String> verifiedIds = FoafSslInterface
				.getVerifiedWebIds(req);
		log.info("Authenticated IDs: " + verifiedIds.toString());

		if (verifiedIds.size() == 0) {
			sendError(req, resp, "Must be authenticated to update data");
			return;
		}

		// step 1: collect underpants (load graphs and attach listener)
		GraphStore graphStore = GraphStoreFactory.create();
		NamedGraphStore ngs = new NamedGraphStore();
		ngs.setConfig(getConfig());
		List<GraphDataListener> gdlList = new ArrayList<GraphDataListener>();

		Collection<String> datasetUris = params.getDatasetUris();
		for (String updateUri : datasetUris) {
			// load or create requested model from store
			Model dbModel = ngs.loadOrCreateModel(updateUri, verifiedIds);
			GraphDataListener gdl = new GraphDataListener(updateUri, dbModel);

			// add graph to temp. graph store
			graphStore.addGraph(gdl.node, gdl.graph);
			gdlList.add(gdl);
		}

		// step 2: ? (execute update)
		UpdateAction.execute(params.getQuery(), graphStore);

		// step 3: profit (check all graphs if changes were allowed)
		for (GraphDataListener gdl : gdlList) {
			String updateUri = gdl.uri;
			// check graph owner, grant full access if user is owner
			DatasetOwnerStore dos = new DatasetOwnerStore(updateUri);
			if (dos.isOwner(updateUri, verifiedIds)) {
				// save changes without further questions
				resp.setStatus(200);
				log.info("Owner update to <" + updateUri + ">");
			} else {
				// load policy
				Collection<String> ruleList = new ArrayList<String>();
				ruleList.add(new RuleStore(updateUri).getRuleStr());
				// check changes with policy
				RuleQuery rq = new RuleQuery(RuleQuery.UPDATE_ACTION, gdl.data,
						verifiedIds, ruleList, getConfig().getPrefixMap());
				// send error if anything goes wrong
				if (!rq.checkUpdates(gdl.ugl.getChangesModel())) {
					sendError(req, resp, "Update to <" + updateUri
							+ "> not authorized. Sorry.");
					ngs.closeConnection();
					ngs.closeGraphSet();
					return;
				}
				log.info("Approved update to <" + updateUri + ">");

			}
			ngs.saveGraph(updateUri, graphStore.getGraph(gdl.node));
		}

		// close persistence
		try {
			ngs.closeConnection();
			ngs.closeGraphSet();
		} catch (Exception e) {
			log.warning("Failed to close persistence layer");
		}
		// finish
		try {
			resp.getWriter().write("OK");
			resp.flushBuffer();
		} catch (IOException e) {
			log.warning("Flushing the response buffer for request "
					+ req.hashCode() + " caused a " + Utils.className(e) + ": "
					+ e.getMessage());
		}
	}
}

class GraphDataListener {
	public Node node;
	public String uri;
	public Model data;
	public Graph graph;
	public UpdateGraphListener ugl;

	public GraphDataListener(String aUri, Model someData) {
		uri = aUri;
		data = someData;
		node = Node.createURI(uri);

		graph = data.getGraph();
		Collection<Triple> originalTriples = graph.find(null, null, null)
				.toList();
		ugl = new UpdateGraphListener(originalTriples);
		graph.getEventManager().register(ugl);
	}
}