package org.pelds.store;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;

import org.pelds.rules.RuleException;

/**
 * API for graph ownership storage
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 * @see DbInterface
 */
public class DatasetOwnerStore extends DbInterface {
	private String dataset;
	private Collection<String> owners;

	/**
	 * Constructor, can be used to create a new ownership relation entry
	 * 
	 * @param datasetUri
	 *            named graph identifier in form of an URI
	 * @param someOwners
	 *            Web IDs of the users with rights to that dataset
	 */
	public DatasetOwnerStore(String datasetUri, Collection<String> someOwners) {
		dataset = datasetUri;
		owners = someOwners;
	}

	/**
	 * Constructor, used to get dataset ownership relation entries
	 * 
	 * @param datasetUri
	 *            named graph identifier in form of an URI
	 */
	public DatasetOwnerStore(String datasetUri) {
		dataset = datasetUri;
	}

	/**
	 * Get the Web IDs of all users with owner rights to a given dataset
	 * 
	 * @param datasetUri
	 *            named graph identifier in form of an URI
	 * @return Collection of Web IDs
	 */
	public Collection<String> getOwnersForDataset(String datasetUri) {
		Collection<String> owners = new ArrayList<String>();
		try {
			PreparedStatement dsStmt = getConnection().prepareStatement(
					"SELECT webid FROM owners WHERE dataset=?");
			dsStmt.setString(1, datasetUri);
			ResultSet rs = dsStmt.executeQuery();
			while (rs.next()) {
				owners.add(rs.getString("webid"));
			}
		} catch (SQLException e) {
			throw new RuleException("SQL Error: " + e.getMessage());
		}
		return owners;
	}

	/**
	 * Stores the current ownership relation entry in the database backend
	 */
	public void save() {
		delete();
		try {
			for (String owner : owners) {
				PreparedStatement ruleStmt = getConnection().prepareStatement(
						"INSERT INTO owners (webid,dataset) VALUES (?,?);");
				ruleStmt.setString(1, owner);
				ruleStmt.setString(2, dataset);
				ruleStmt.executeUpdate();
			}

		} catch (SQLException e) {
			throw new RuleException("SQL Error: " + e.getMessage());

		}
	}

	/**
	 * Deletes the current ownership relation entry from the database backend
	 */
	public void delete() {
		try {
			PreparedStatement ruleStmt = getConnection().prepareStatement(
					"DELETE FROM owners WHERE dataset=?");
			ruleStmt.setString(1, dataset);
			ruleStmt.executeUpdate();
		} catch (SQLException e) {
			throw new RuleException("SQL Error: " + e.getMessage());
		}
	}

	/**
	 * Check if one of the given users identified by their Web ID has a
	 * ownership entry in the database for a given dataset
	 * 
	 * @param dataset
	 *            named graph identifier in form of an URI
	 * @param verifiedIds
	 *            Web IDs of the users to check
	 * @return true if one of the users has owner rights to the dataset
	 */
	public boolean isOwner(String dataset, Collection<String> verifiedIds) {
		Collection<String> owners = getOwnersForDataset(dataset);
		for (String webID : verifiedIds) {
			if (owners.contains(webID))
				return true;
		}
		return false;
	}

	public String getDataset() {
		return dataset;
	}

	public void setDataset(String dataset) {
		this.dataset = dataset;
	}

	public Collection<String> getOwners() {
		return owners;
	}

	public void setOwners(Collection<String> someOwners) {
		this.owners = someOwners;
	}

}
