package org.pelds.store;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.logging.Logger;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

import de.fuberlin.wiwiss.ng4j.NamedGraph;
import de.fuberlin.wiwiss.ng4j.NamedGraphSet;
import de.fuberlin.wiwiss.ng4j.Quad;
import de.fuberlin.wiwiss.ng4j.db.NamedGraphSetDB;

/**
 * Named graph storage, uses NG4J. Provides methods for graph creation, loading,
 * storing, locating
 * 
 * @author Hannes Muehleisen (hannes@muehleisen.org)
 */
public class NamedGraphStore extends DbInterface {

	public static final String TABLE_PREFIX = "data";
	static private Logger log = Logger.getLogger("NamedGraphStore");

	private NamedGraphSet graphset = null;

	/**
	 * Create a NG4J graph set with the current database connection
	 * 
	 * @return a NG4J graph set
	 */
	public NamedGraphSet getGraphSet() {
		if (graphset == null) {
			graphset = new NamedGraphSetDB(getConnection(), TABLE_PREFIX);
		}
		return graphset;
	}

	/**
	 * Retrieves a RDF graph / model from the graph store by the graph name. If
	 * a graph with that name does not exist, it gets created.
	 * 
	 * @param uri
	 *            named graph identifier in form of an URI
	 * @param authenticatedIds
	 *            Web IDs of the authenticated client
	 * @return a RDF graph / model
	 */
	public Model loadOrCreateModel(String uri,
			Collection<String> authenticatedIds) {
		Graph g = getGraphSet().getGraph(uri);
		Model m = ModelFactory.createDefaultModel();
		// if the graph does not exist, return an empty model
		if (g == null) {
			// but create the graph and set the authenticated ids as owners
			if (authenticatedIds != null && authenticatedIds.size() > 0) {
				getGraphSet().createGraph(uri);
				DatasetOwnerStore dos = new DatasetOwnerStore(uri,
						authenticatedIds);
				dos.save();
				log.info("Created new Dataset <" + uri + ">");
			}
			return m;
		}
		ExtendedIterator<Triple> tripleList = g.find(null, null, null);
		while (tripleList.hasNext()) {
			m.add(m.asStatement(tripleList.next()));
		}
		return m;
	}

	/**
	 * Retrieves a RDF graph / model from the graph store by the graph name.
	 * 
	 * @param uri
	 *            named graph identifier in form of an URI
	 * @return a RDF graph / model
	 */
	public Model loadModel(String uri) {
		return loadOrCreateModel(uri, null);
	}

	/**
	 * Stores a graph to the database
	 * 
	 * @param uri
	 *            named graph identifier in form of an URI
	 * @param m
	 *            changed RDF graph
	 */
	public void saveGraph(String uri, Graph m) {
		getGraphSet().removeGraph(uri);
		NamedGraph nm = getGraphSet().createGraph(uri);
		ExtendedIterator<Triple> it = m.find(null, null, null);
		while (it.hasNext()) {
			nm.add(it.next());
		}

	}

	/**
	 * Close NG4Js database connection
	 */
	public void closeGraphSet() {
		/*
		 * if (graphset != null) { graphset.close(); graphset = null; }
		 */
		// disabled for now
	}

	/**
	 * Find all datasets mentioning either mentioning the given resource as name
	 * or as graph identifier.
	 * 
	 * @param resUri
	 *            a RDF resource URI
	 * @return Collection of RDF graph identifiers
	 */
	@SuppressWarnings("unchecked")
	public Collection<String> findDatasetForResource(String resUri) {
		Collection<String> retList = new ArrayList<String>();
		Resource res = ModelFactory.createDefaultModel().createResource(resUri);
		Iterator<Quad> qIt = (Iterator<Quad>) getGraphSet().findQuads(null,
				res.asNode(), null, null);
		while (qIt.hasNext()) {
			String graphName = qIt.next().getGraphName().toString();
			if (!retList.contains(graphName)) {
				retList.add(graphName);
			}
		}

		qIt = (Iterator<Quad>) getGraphSet().findQuads(res.asNode(), null,
				null, null);
		while (qIt.hasNext()) {
			String graphName = qIt.next().getGraphName().toString();
			if (!retList.contains(graphName)) {
				retList.add(graphName);
			}
		}
		return retList;
	}
}
