package org.pelds.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.net.URL;
import java.util.ArrayList;

import org.junit.Before;
import org.junit.Test;
import org.pelds.rules.RuleQuery;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.shared.PrefixMapping;

public class RuleQueryTests {
	private Model data;

	@Before
	public void init() {
		data = ModelFactory.createDefaultModel();
		URL file = getClass().getResource("test.ttl");
		data.read(file.toString(), "TTL");
		data.write(System.err, "TTL");
	}

	// empty result with no rules
	@Test
	public void queryNoRules() {
		Model m = query("");
		assertEquals(m.size(), 0);
	}

	// complete result with liberal rule
	@Test
	public void queryPublicRules() {
		Model m = query("PublicRule: QueryAction(?a) => permit_triple(*,*,*);");
		assertTrue(m.containsAll(data));
	}
	
	// complete result with prefixing
	@Test
	public void queryPublicPrefixRules() {
		Model m = query("PublicRule: pelds:QueryAction(?a) => pelds:permit_triple(*,*,*);");
		assertTrue(m.containsAll(data));
	}

	// partial result for this user by triple
	@Test
	public void queryUserTripRules() {
		Model m = query("UserTripRule: QueryAction(?a) && actor(?a,p://user1) => permit_triple(p://whatever,*,*);");
		assertTrue(m.containsResource(ResourceFactory
				.createResource("p://whatever"))
				&& !m.containsResource(ResourceFactory
						.createResource("p://whatever2")));
	}

	// partial result for this user by resource
	@Test
	public void queryUserResRules() {
		Model m = query("UserTripRule: QueryAction(?a) && actor(?a,p://user1) => permit_resource(p://whatever);");
		assertTrue(m.containsResource(ResourceFactory
				.createResource("p://whatever"))
				&& !m.containsResource(ResourceFactory
						.createResource("p://whereever")));
	}

	
	// partial result for this user by instance
	@Test
	public void queryUserInstRules() {
		Model m = query("UserInstRule: QueryAction(?a) && actor(?a,p://user1) => permit_instance(ex:Something);");
		assertTrue(m.containsResource(ResourceFactory
				.createResource("p://whatever"))
				&& !m.containsResource(ResourceFactory
						.createResource("p://whereever")));
	}

	// partial result for this user by instance, with additional consequence, no prefix mapping
	@Test
	public void queryUserConNoPMRules() {
		Model m = query("SwrlConRule: p://ont#Otherthing(?t) => p://ont#Something(?t); UserInstRule: pelds:QueryAction(?a) => permit_instance(p://ont#Something);");
		assertTrue(m.containsResource(ResourceFactory
						.createResource("p://whereever")));
	}
	
	// partial result for this user by instance, with additional consequence
	@Test
	public void queryUserConRules() {
		Model m = query("SwrlConRule: ex:Otherthing(?t) => ex:Something(?t); UserInstRule: QueryAction(?a) => permit_instance(ex:Something);");
		assertTrue(m.containsResource(ResourceFactory
						.createResource("p://whereever")));
	}

	// partial result for this user by instance, with sameas
	@Test
	public void queryUserSameasRules() {
		Model m = query("SwrlSameAsRule: sameAs(?t,p://whereever) => ex:Something(?t); UserInstRule: QueryAction(?a) => permit_instance(ex:Something);");
		assertTrue(m.containsResource(ResourceFactory
						.createResource("p://whereever")));
	}

	// partial result for this user by instance, with differentFrom
	@Test
	public void queryUserDifffromRules() {
		Model m = query("SwrlDiffFromRule: ex:Otherthing(?t) && differentFrom(?t,p://whatever) => ex:Something(?t); UserInstRule: QueryAction(?a) => permit_instance(ex:Something);");
		assertTrue(m.containsResource(ResourceFactory
				.createResource("p://whereever")));
	}

	// partial result for this user by instance, with datatype property
	@Test
	public void queryUserDatatypeRules() {
		Model m = query("SwrlDatatypeRule: ex:id(?t,\"213\") => ex:Something(?t); UserInstRule: QueryAction(?a) => permit_instance(ex:Something);");
		assertTrue(m.containsResource(ResourceFactory
				.createResource("p://whereever")));
	}

	// partial result for this user by instance, with builtin function
	@Test
	public void queryUserBuiltinRules() {
		Model m = query("SwrlBuiltinRule: ex:Otherthing(?t) && ex:id(?t,?i) && swrlb:startsWith(?i,\"213\") => ex:Something(?t); UserInstRule: QueryAction(?a) => permit_instance(ex:Something);");
		assertTrue(m.containsResource(ResourceFactory
				.createResource("p://whereever")));
	}
	
	// helper
	private Model query(String rule) {
		ArrayList<String> identities = new ArrayList<String>();
		identities.add("p://user1");
		identities.add("p://user2");
		ArrayList<String> rules = new ArrayList<String>();
		rules.add(rule);
		PrefixMapping pm = PrefixMapping.Factory.create();
		pm.setNsPrefixes(PrefixMapping.Extended);
		pm.setNsPrefix("ex", "p://ont#");
		RuleQuery rq = new RuleQuery(RuleQuery.QUERY_ACTION, data, identities,
				rules, pm);
		rq.setVerbose(true);
		return rq.getCheckedModel();
	}
}